#include "Convexhull2D.h"

#include <algorithm>
#include <limits>

using namespace std;
using namespace gmtl;

float ccw(const Vec2f& p, const Vec2f& q, const Vec2f& r) {
	return (q[0] - p[0]) * (r[1] - p[1]) - (r[0] - p[0]) * (q[1] - p[1]);
}

#define areaTriangle(p, q, r) (ccw(p, q, r) * 0.5f)

Convexhull2D::Convexhull2D(gmtl::Vec2f* a, size_t n) : iSize(0) {
	construct(a, n);
}

Convexhull2D::Convexhull2D() : iSize(0) {

}

Convexhull2D::Convexhull2D(const Convexhull2D& ch) : iSize(0) {
	*this = ch;
}

Convexhull2D::~Convexhull2D() {

}

float area(std::vector<gmtl::Vec2f>& v) {
	float fArea = 0.0f;
	for(int i = 1; i < v.size() - 1; i++) {
		fArea += static_cast<float>(fabs(areaTriangle(v[0], v[i], v[i + 1])));
	}

	return fArea;
}

int indexcompare(const void* a, const void* b) {
	Convexhull2D::Vectorindex va = *(Convexhull2D::Vectorindex*)a;
	Convexhull2D::Vectorindex vb = *(Convexhull2D::Vectorindex*)b;

	if(fabs(va.pv[0] - vb.pv[0]) < numeric_limits<float>::epsilon()) {
		if(va.pv[1] >= vb.pv[1])
			return 1;
		else
			return  0;
	} else if(va.pv[0] >= vb.pv[0]) {
		return  1;
	}

	return -1;
}

struct VectorindexCompare {
	int operator() (const Convexhull2D::Vectorindex& va, const Convexhull2D::Vectorindex& vb) const {
		if(fabs(va.pv[0] - vb.pv[0]) < numeric_limits<float>::epsilon()) {
			if(va.pv[1] >= vb.pv[1])
				return 1;
			else
				return  0;
		} else if(va.pv[0] >= vb.pv[0]) {
			return 1;
		}

		return 0;
	}
};

void Convexhull2D::construct(gmtl::Vec2f* points, size_t n) {
	vector<Convexhull2D::Vectorindex> sortedIndex(n);
	vector<Vectorindex> chpoints;
	
	size_t nsize;
		
	int i, j, k1, k2;

	float f;
	
	for(i = 0; i < n; i++) {
		sortedIndex[i].i = i;
		sortedIndex[i].pv = points[i];
	}
	
	sort(sortedIndex.begin(), sortedIndex.end(), VectorindexCompare());
	
	chpoints.resize(n + 1);
		
	sortedIndex[0].flag = true;
	sortedIndex[1].flag = false;

	chpoints[0] = sortedIndex[0];
	chpoints[1] = sortedIndex[1];

	chpoints[0].pi = &sortedIndex[0];
	chpoints[1].pi = &sortedIndex[1];

	k1 = 2;

	for(i = 2; i < n; i++) {
		for(j = k1 - 1; j > 0; j--) {				
			f = ccw(chpoints[j - 1].pv, points[sortedIndex[i].i], chpoints[j].pv); 
			if(fabs(f) < numeric_limits<float>::epsilon() || f >= 0.0f) {
				chpoints[--k1].pi->flag = true;
			} else
				break;
		}
		sortedIndex[i].flag = false;
		chpoints[k1] = sortedIndex[i];
		chpoints[k1++].pi = &sortedIndex[i];
	}

	k2 = 0;
	
	for(i = n - 2; i >= 0; i--) {
		if(sortedIndex[i].flag) {
			for(j = k2 - 1; j >= 0; j--) {				
				f = ccw(chpoints[k1 + j - 1].pv, points[sortedIndex[i].i], chpoints[k1 + j].pv);
				if(fabs(f) < numeric_limits<float>::epsilon() || f >= 0.0f)
					k2--;
				else
					break;
			}
			chpoints[k1 + k2++] = sortedIndex[i];		
		}
	} 

	nsize = k1 + k2 - 1;

	chpoints.resize(nsize);

	vertices.resize(iSize = nsize);
	
	for(i = 0; i < iSize; i++) {
		vertices[i] = chpoints[i].pv;
	}
}